@@ -43,6 +43,8 @@ static GHashTable *rfkill_list;
 
 static bool global_offlinemode;
 
+static char *global_regdom = NULL;
+
 struct connman_rfkill {
 	unsigned int index;
 	enum connman_service_type type;
@@ -333,8 +335,14 @@ int connman_technology_set_regdom(const char *alpha2)
 					driver->set_regdom(technology, alpha2);
 			}
 		}
+
+		/* Save regdom for this technology */
+		connman_technology_regdom_notify(technology, alpha2);
 	}
 
+	g_free(global_regdom);
+	global_regdom = g_strdup(alpha2);
+
 	return 0;
 }
 
@@ -354,6 +362,22 @@ static struct connman_technology *technology_find(enum connman_service_type type
 	return NULL;
 }
 
+const char *__connman_technology_get_regdom(enum connman_service_type type)
+{
+	struct connman_technology *technology;
+
+	DBG("type %d/%s", type, get_name(type));
+
+	technology = technology_find(type);
+	if (!technology)
+		return NULL;
+
+	if (technology->regdom)
+		return technology->regdom;
+
+	return global_regdom;
+}
+
 enum connman_service_type connman_technology_get_type
 				(struct connman_technology *technology)
 {
@@ -405,19 +429,26 @@ static void technology_load(struct connman_technology *technology)
 	GKeyFile *keyfile;
 	gchar *identifier;
 	GError *error = NULL;
-	bool enable, need_saving = false;
+	bool enable, need_saving = false, enable_persistent = false;
 	char *enc;
+	int i;
+	unsigned int *enabled_types =
+		connman_setting_get_uint_list("DefaultEnabledTechnologies");
 
 	DBG("technology %p", technology);
 
+	/* Determine which technology types are enabled by default */
+	for (i = 0; enabled_types && enabled_types[i] != 0; i++) {
+		if (technology->type == enabled_types[i]) {
+			enable_persistent = true;
+			break;
+		}
+	}
+
 	keyfile = __connman_storage_load_global();
 	/* Fallback on disabling technology if file not found. */
 	if (!keyfile) {
-		if (technology->type == CONNMAN_SERVICE_TYPE_ETHERNET)
-			/* We enable ethernet by default */
-			technology->enable_persistent = true;
-		else
-			technology->enable_persistent = false;
+		technology->enable_persistent = enable_persistent;
 		return;
 	}
 
@@ -429,10 +460,7 @@ static void technology_load(struct connman_technology *technology)
 	if (!error)
 		technology->enable_persistent = enable;
 	else {
-		if (technology->type == CONNMAN_SERVICE_TYPE_ETHERNET)
-			technology->enable_persistent = true;
-		else
-			technology->enable_persistent = false;
+		technology->enable_persistent = enable_persistent;
 
 		need_saving = true;
 		g_clear_error(&error);
@@ -455,8 +483,10 @@ static void technology_load(struct connman_technology *technology)
 
 	enc = g_key_file_get_string(keyfile,
 				identifier, "Tethering.Passphrase", NULL);
-	if (enc)
+	if (enc) {
 		technology->tethering_passphrase = g_strcompress(enc);
+		g_free(enc);
+	}
 
 	technology->tethering_freq = g_key_file_get_integer(keyfile,
 				identifier, "Tethering.Freq", NULL);
@@ -1283,6 +1313,7 @@ static struct connman_technology *technology_get(enum connman_service_type type)
 	technology_load(technology);
 	technology_list = g_slist_prepend(technology_list, technology);
 	technology->driver_list = tech_drivers;
+	technology->regdom = g_strdup(global_regdom);
 
 	for (list = tech_drivers; list; list = list->next) {
 		driver = list->data;
@@ -1905,4 +1936,6 @@ void __connman_technology_cleanup(void)
 	g_hash_table_destroy(rfkill_list);
 
 	dbus_connection_unref(connection);
+
+	g_free(global_regdom);
 }
